import os
import tkinter

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.image as mpimg

from typing import Dict

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
from scipy.spatial.distance import cdist

from backend import load_map_from_ply_and_txt
from modules.lines_utils import Point
from models.ModelMap import ModelMap
from models.ModelToggles import ModelToggles
from models.ModelPOI import ModelPOI

path_current_dir = os.path.dirname(__file__)
path_base = f'{path_current_dir}/../'


class ControllerMap:
    def __init__(self, master, select_poi_mb, delete_poi_btn):
        self.fig = plt.Figure(figsize=(10, 10), dpi=150)
        self.fig.set_figwidth(10)
        self.fig.set_figheight(4)
        self.ax = self.fig.add_subplot(111)
        self.canvas = FigureCanvasTkAgg(self.fig, master=master)
        self.canvas.get_tk_widget().pack(fill='both', expand='true')
        self.__setup_plot()

        self.model_map = ModelMap()
        self.model_toggles = ModelToggles()
        self.is_snap_to_flight_path = True

        self.plt_lines_by_idx: Dict[int, object] = {}
        self.plt_poi_by_point: Dict[Point, object] = {}

        self.press = False
        self.move = False

        self.tmp_point_original_poi = None

        self.canvas.mpl_connect('button_press_event', self.onpress)
        self.canvas.mpl_connect('button_release_event', self.onrelease)
        self.canvas.mpl_connect('motion_notify_event', self.onmove)

        self.select_poi_mb = select_poi_mb
        self.select_poi_mb.menu = tkinter.Menu(self.select_poi_mb, tearoff=0)
        self.select_poi_mb["menu"] = self.select_poi_mb.menu

        self.delete_poi_btn = delete_poi_btn
        self.delete_poi_btn.configure(command=self.delete_poi)

    def __setup_plot(self):
        self.ax.set_aspect('equal')
        self.ax.invert_yaxis()
        self.ax.invert_xaxis()
        self.ax.axis('off')
        self.fig.patch.set_facecolor('black')
        self.fig.tight_layout()

    def clear_plot(self):
        self.model_map.is_deleted = True
        self.ax.clear()
        self.__setup_plot()
        self.canvas.draw()

    def plot(self):
        # TODO: add filtered pointcloud
        if not self.model_map.is_deleted:
            self.ax.clear()

            # Plot walls
            if self.model_toggles.showWalls and self.model_map.line_start_points_in_pix is not None:
                for idx in range(len(self.model_map.line_start_points_in_pix)):
                    plt_plot, = self.ax.plot(
                        [
                            self.model_map.line_start_points_in_pix[idx, 0],
                            self.model_map.line_end_points_in_pix[idx, 0]
                        ],
                        [
                            self.model_map.line_start_points_in_pix[idx, 1],
                            self.model_map.line_end_points_in_pix[idx, 1]
                        ],
                        c='#0085ff', linewidth=3.0)
                    self.plt_lines_by_idx[idx] = plt_plot

            # Plot original pointcloud
            if self.model_toggles.showRawPointcloud and self.model_map.raw_point_cloud_in_pix is not None:
                x = map(lambda point: point.x, self.model_map.raw_point_cloud_in_pix)
                y = map(lambda point: point.y, self.model_map.raw_point_cloud_in_pix)
                self.ax.scatter(list(x), list(y), s=0.5, c='#ff5656')

            # Plot filtered pointcloud
            if self.model_toggles.showFilteredPointcloud and self.model_map.filtered_point_cloud_in_pix is not None:
                x = map(lambda point: point.x, self.model_map.filtered_point_cloud_in_pix)
                y = map(lambda point: point.y, self.model_map.filtered_point_cloud_in_pix)
                self.ax.scatter(list(x), list(y), s=0.5, c='grey')

            # Plot drone path
            if self.model_toggles.showFligthPath and self.model_map.path_points_in_pix is not None:
                x = map(lambda point: point.x, self.model_map.path_points_in_pix)
                y = map(lambda point: point.y, self.model_map.path_points_in_pix)
                self.ax.plot(list(x), list(y), c='#00e224', linewidth=2)

            # Plot POIs
            if self.model_toggles.showAllPOIs:
                # TODO fix ineffective POI redrawing (canvas.draw() is called after each icon is added to ax)
                # TODO plot 2 types of POI (active, inactive) and drone
                if len(self.model_map.pois_in_pix) != 0:
                    for poi in self.model_map.pois_in_pix:
                        self.plot_poi(poi, 'target')
            else:
                if len(self.model_map.pois_in_pix) != 0:
                    for poi in self.model_map.pois_in_pix:
                        if poi.is_selected:
                            self.plot_poi(poi, 'target')

            self.__setup_plot()
            self.canvas.draw()

    def plot_poi(self, poi, poi_type: str):
        # TODO change pictures
        if poi.is_selected:
            img = mpimg.imread(f'{path_base}/assets/poi-selected.png')
        else:
            img = mpimg.imread(f'{path_base}/assets/poi-unselected.png')
        img = OffsetImage(img, zoom=0.3)
        img = AnnotationBbox(img, (poi.point.x, poi.point.y), frameon=False)

        self.plt_poi_by_point[poi.point] = img
        self.ax.add_artist(img)
        self.canvas.draw()

    def update_toggles(self, modelToggles):
        self.model_toggles = modelToggles
        self.plot()

    def load_map_from_ply(self, ply_filepath, path_filepath=None, use_45deg_lines=False):
        map = load_map_from_ply_and_txt(ply_filepath, path_filepath, use_45deg_lines)
        map.rotate_and_scale_data()

        self.model_map.pois_in_pix.clear()
        self.model_map.raw_point_cloud_in_pix = map.point_cloud_in_pix
        self.model_map.filtered_point_cloud_in_pix = map.filtered_point_cloud_in_pix
        self.model_map.path_points_in_pix = map.path_points
        self.model_map.pix_to_meters = map.pix_to_meters
        self.model_map.meters_to_pix = map.meters_to_pix
        self.model_map.line_start_points_in_pix = map.line_start_points_in_pix
        self.model_map.line_end_points_in_pix = map.line_end_points_in_pix
        self.model_map.is_deleted = False

    def add_poi(self, point):
        if len(self.model_map.pois_in_pix):
            poi_id = self.model_map.pois_in_pix[-1].id + 1
        else:
            poi_id = 0

        self.model_map.pois_in_pix.append(
            ModelPOI(id=poi_id,
                     name=f'POI_{poi_id}',
                     point=point,
                     is_selected=False))
        self.select_poi_mb.menu.add_command(label=f'POI_{poi_id}',
                                            command=lambda: self.select_poi(poi_id, replot=True))
        return poi_id

    def delete_poi(self):
        if not self.model_map.is_deleted and len(self.model_map.pois_in_pix):
            for idx, poi in enumerate(self.model_map.pois_in_pix):
                if poi.is_selected:
                    self.model_map.pois_in_pix.pop(idx)
                    self.select_poi_mb.menu.delete(f'POI_{poi.id}')
                    self.select_poi(replot=True)

    def select_poi(self, poi_id=-1, replot=False):
        if len(self.model_map.pois_in_pix) == 1:
            self.model_map.pois_in_pix[0].is_selected = True
        else:
            for poi in self.model_map.pois_in_pix:
                if poi.is_selected:
                    poi.is_selected = False
                if poi.id == poi_id:
                    poi.is_selected = True

        if poi_id==-1:
            self.select_poi_mb.configure(text=f'Select POI')
        else:
            self.select_poi_mb.configure(text=f'POI_{poi_id}')

        if replot:
            self.plot()

    def onpress(self, event):
        try:
            if self.model_toggles.showAllPOIs and not self.model_map.is_deleted:
                if not self.move and event.dblclick:
                    point_poi = Point(event.xdata, event.ydata)

                    if self.model_toggles.snapToFlightPath:
                        np_points_flight_path = np.array([[it.x, it.y] for it in self.model_map.path_points_in_pix])
                        np_point = np.array([[point_poi.x, point_poi.y]])
                        dists = cdist(np_point, np_points_flight_path)
                        idx_min_dist = dists.squeeze().argmin()
                        point_poi.x = np_points_flight_path[idx_min_dist, 0]
                        point_poi.y = np_points_flight_path[idx_min_dist, 1]

                    poi_id = self.add_poi(point_poi)
                    self.select_poi(poi_id)

                    point_original = np.matmul(self.model_map.pix_to_meters,
                                               np.array([event.xdata, event.ydata, 1]))
                    self.tmp_point_original_poi = point_original

                    print(f'x: {point_original[0]}, y:{point_original[1]}')
                    self.plot()
        except Exception as e:
            print(e) # TODO debug panel

    def onmove(self, event):
        # TODO: onmove event
        # print('onmove', event)
        if self.press:
            self.move = True

    def onrelease(self, event):
        # TODO: onrelease event
        # print('onrelease', event)
        self.press = False
        self.move = False
